import 'dotenv/config';
import express from 'express';
import { Express, Request, Response, NextFunction } from 'express'
import { access, copyFile, unlink, mkdir, readFile, rm } from 'fs/promises';
import { tmpdir } from 'os';
import { join, basename, resolve, extname } from 'path';
import bodyParser from 'body-parser';
import multer, { FileFilterCallback } from 'multer';
import { v4 as uuid } from 'uuid';
import getType from 'mime';
import { createLog } from './log'
import type { Logger } from 'winston';
import { envInt, envString } from './env';
import { Shell } from './shell';

const PORT : number = envInt('PORT', 7474);
const BIN : string = envString('BIN', 'openpose');
const CWD : string = envString('CWD', '.');

const log : Logger = createLog('openpose_api');
const app : Express = express();
const tmp : string = tmpdir();

const storage = multer.diskStorage({    
	destination: function (req : Request, file : Express.Multer.File, cb : Function) {
        cb(null, tmp)
    },
    filename: function (req : Request, file : Express.Multer.File, cb : Function) {
        cb(null, `${+new Date()}_${file.originalname}`)
    }
});

async function exists (path : string) : Promise<boolean> {  
	try {
		await access(path);
		return true;
	} catch {
		return false;
	}
}

app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: true }));

const uploadImage : any = multer({ storage });

app.get('/', (req: Request, res: Response, next : NextFunction) => {
	return res.send('/');
});

//./build/examples/openpose/openpose.bin -v 1 -image_dir ./examples/media/ -write_images ./output -write_json ./json --face --hand
app.post('/openpose', uploadImage.single('image'), async (req: Request, res: Response, next : NextFunction) => {
	const id : string = uuid();
	const dir : string = join(tmp, id);
	const image_dir : string = join(dir, 'image');
	const json_dir : string = join(dir, 'json');
	const json_file : string = join(json_dir, 'openpose_keypoints.json');
	let json_exists : boolean = false;
	let openpose_str : string = '{}';
	let output : any;
	let image_file : string;
	let image_ext : string;
	let sh : Shell;
	//
	const args : string[] = [
		BIN,
		'--render_pose', '0',
		'-display', '0',
		'-image_dir', image_dir,
		'-write_json', json_dir
	];

	if (typeof req.file === 'undefined') {
		return next(new Error('No image uploaded'));
	}

	image_ext = extname(basename(req.file.path));
	image_file = join(image_dir, `openpose${image_ext}`);

	if (typeof req.body.face !== 'undefined' && req.body.face === 'true') {
		args.push('--face');
	}

	if (typeof req.body.hand !== 'undefined' && req.body.hand === 'true') {
		args.push('--hand');
	}

	try {
		await mkdir(dir); 
	} catch (err) {
		log.error(`Error making directory ${dir}`, err);
		return next(new Error('Error making dir'));
	}

	try {
		await mkdir(image_dir);
	} catch (err) {
		log.error(`Error making directory ${image_dir}`, err);
		return next(new Error('Error making image_dir'));
	}

	try {
		await mkdir(json_dir);
	} catch (err) {
		log.error(`Error making directory ${image_dir}`, err);
		return next(new Error('Error making image_dir'));
	}

	try {
		await copyFile(req.file.path, image_file);
	} catch (err) {
		log.error(`Error copying file ${req.file.path}`, err);
		return next(new Error('Error copying temp image'));
	}

	sh = new Shell(args, CWD, null, null, null, true);

	try {
		await sh.execute();
	} catch (err) {
		log.error(`Error executing ${args.join(' ')}`, err);
		return next(new Error('Error running openpose'));
	}

	try {
		json_exists = await exists(json_file);
	} catch (err) {
		log.error(`Error checking json ${json_file} exists`, err);
		return next(new Error('Error checking json exists'));
	}

	if (json_exists) {
		try {
			openpose_str = await readFile(json_file, 'utf8');
		} catch (err) {
			log.error(`Error reading json file ${json_file}`, err);
			return next(new Error('Error reading json file'));
		}
	} else {
		log.error(`The json file ${json_file} does not exist`);
		return next(new Error('Error running openpose'));
	}

	try {
		output = JSON.parse(openpose_str);
	} catch (err) {
		log.error(`Error parsing openpose json`, err);
		return next(new Error('Error parsing openpose json'));
	}

	output.image = req.file.originalname;

	try {
		await rm(dir, { recursive: true });
	} catch (err) {
		log.error(`Error unlinking ${dir}`, err);
		return next(new Error('Error unlinking temp dir'));
	}

	return res.json(output);
});

app.listen(PORT, () => {
	log.info(`Openpose API server running on port ${PORT}`);
	log.info(`Using openpose binary ${BIN}`);
});